@@ -6,42 +6,40 @@ from email.header import Header
 from email.mime.application import MIMEApplication
 from email.mime.multipart import MIMEMultipart
 from email.mime.nonmultipart import MIMENonMultipart
-from email.utils import formataddr, getaddresses
+from email.utils import getaddresses
+
+from sql.operators import Concat
+from genshi.template import TextTemplate
 
 from trytond.config import config
+from trytond.i18n import gettext
 from trytond.model import ModelView, ModelSQL, fields
 from trytond.pool import Pool
 from trytond.pyson import Eval
 from trytond.report import get_email
 from trytond.sendmail import sendmail_transactional, SMTPDataManager
 from trytond.transaction import Transaction
+from .exceptions import TemplateError
 
-_EMAIL_MODELS = [
-    'res.user',
-    'party.party',
-    'web.user',
-    'company.employee',
-    'commission.agent']
-
-
-def _formataddr(name, email):
-    if name:
-        name = str(Header(name, 'utf-8'))
-    return formataddr((name, email))
+from trytond.ir.resource import ResourceAccessMixin
 
 
 class Email(ModelSQL, ModelView):
-    "Email Notitification"
+    "Email Notification"
     __name__ = 'notification.email'
 
     from_ = fields.Char(
         "From",
         help="Leave empty for the value defined in the configuration file.")
+    subject = fields.Char(
+        "Subject", translate=True,
+        help="The Genshi syntax can be used "
+        "with 'record' in the evaluation context.\n"
+        "If empty the report name will be used.")
     recipients = fields.Many2One(
         'ir.model.field', "Recipients",
         domain=[
             ('model.model', '=', Eval('model')),
-            ('relation', 'in', _EMAIL_MODELS),
             ],
         depends=['model'],
         help="The field that contains the recipient(s).")
@@ -59,7 +57,6 @@ class Email(ModelSQL, ModelView):
         'ir.model.field', "Secondary Recipients",
         domain=[
             ('model.model', '=', Eval('model')),
-            ('relation', 'in', _EMAIL_MODELS),
             ],
         depends=['model'],
         help="The field that contains the secondary recipient(s).")
@@ -77,7 +74,6 @@ class Email(ModelSQL, ModelView):
         'ir.model.field', "Hidden Recipients",
         domain=[
             ('model.model', '=', Eval('model')),
-            ('relation', 'in', _EMAIL_MODELS),
             ],
         depends=['model'],
         help="The field that contains the hidden recipient(s).")
@@ -118,6 +114,25 @@ class Email(ModelSQL, ModelView):
     model = fields.Function(
         fields.Char("Model"), 'on_change_with_model', searcher='search_model')
 
+    @classmethod
+    def __setup__(cls):
+        pool = Pool()
+        EmailTemplate = pool.get('ir.email.template')
+        super().__setup__()
+        for field in [
+                'recipients',
+                'recipients_secondary',
+                'recipients_hidden',
+                ]:
+            field = getattr(cls, field)
+            field.domain.append(['OR',
+                    ('relation', 'in', EmailTemplate.email_models()),
+                    [
+                        ('model.model', 'in', EmailTemplate.email_models()),
+                        ('name', '=', 'id'),
+                        ],
+                    ])
+
     def get_rec_name(self, name):
         return self.content.rec_name
 
@@ -143,96 +158,16 @@ class Email(ModelSQL, ModelView):
     def search_model(cls, name, clause):
         return [('content.model',) + tuple(clause[1:])]
 
-    def _get_address(self, record):
-        pool = Pool()
-        User = pool.get('res.user')
-        try:
-            Party = pool.get('party.party')
-        except KeyError:
-            Party = None
-        try:
-            WebUser = pool.get('web.user')
-        except KeyError:
-            WebUser = None
-        try:
-            Employee = pool.get('company.employee')
-        except KeyError:
-            Employee = None
-        try:
-            Agent = pool.get('commission.agent')
-        except KeyError:
-            Agent = None
-        if isinstance(record, User) and record.email:
-            return _formataddr(record.rec_name, record.email)
-        elif Party and isinstance(record, Party):
-            contact = record.contact_mechanism_get(
-                'email', usage=self.contact_mechanism)
-            if contact and contact.email:
-                return _formataddr(
-                    contact.name or record.rec_name, contact.email)
-        elif WebUser and isinstance(record, WebUser):
-            name = None
-            if record.party:
-                name = record.party.rec_name
-            return _formataddr(name, record.email)
-        elif Employee and isinstance(record, Employee):
-            return self._get_address(record.party)
-        elif Agent and isinstance(record, Agent):
-            return self._get_address(record.party)
-
     def _get_addresses(self, value):
-        if isinstance(value, (list, tuple)):
-            addresses = (self._get_address(v) for v in value)
-        else:
-            addresses = [self._get_address(value)]
-        return [a for a in addresses if a]
-
-    def _get_language(self, record):
         pool = Pool()
-        Configuration = pool.get('ir.configuration')
-        User = pool.get('res.user')
-        Lang = pool.get('ir.lang')
-        try:
-            Party = pool.get('party.party')
-        except KeyError:
-            Party = None
-        try:
-            WebUser = pool.get('web.user')
-        except KeyError:
-            WebUser = None
-        try:
-            Employee = pool.get('company.employee')
-        except KeyError:
-            Employee = None
-        try:
-            Agent = pool.get('commission.agent')
-        except KeyError:
-            Agent = None
-        if isinstance(record, User):
-            if record.language:
-                return record.language
-        elif Party and isinstance(record, Party):
-            if record.lang:
-                return record.lang
-        elif WebUser and isinstance(record, WebUser):
-            if record.party and record.party.lang:
-                return record.party.lang
-        elif Employee and isinstance(record, Employee):
-            if record.party.lang:
-                return record.party.lang
-        elif Agent and isinstance(record, Agent):
-            if record.party.lang:
-                return record.party.lang
-        lang, = Lang.search([
-                ('code', '=', Configuration.get_language()),
-                ], limit=1)
-        return lang
+        EmailTemplate = pool.get('ir.email.template')
+        with Transaction().set_context(usage=self.contact_mechanism):
+            return EmailTemplate.get_addresses(value)
 
     def _get_languages(self, value):
-        if isinstance(value, (list, tuple)):
-            return {self._get_language(v) for v in value}
-        else:
-            return {self._get_language(value)}
+        pool = Pool()
+        EmailTemplate = pool.get('ir.email.template')
+        return EmailTemplate.get_languages(value)
 
     def get_email(self, record, from_, to, cc, bcc, languages):
         pool = Pool()
@@ -240,11 +175,17 @@ class Email(ModelSQL, ModelView):
 
         # TODO order languages to get default as last one for title
         content, title = get_email(self.content, record, languages)
+        language = list(languages)[-1]
+        if self.subject:
+            with Transaction().set_context(language=language.code):
+                notification = self.__class__(self.id)
+                title = (TextTemplate(notification.subject)
+                    .generate(record=record)
+                    .render())
 
         if self.attachments:
             msg = MIMEMultipart('mixed')
             msg.attach(content)
-            language = list(languages)[-1]
             for report in self.attachments:
                 msg.attach(Attachment.get_mime(report, record, language.code))
         else:
@@ -262,6 +203,8 @@ class Email(ModelSQL, ModelView):
             'recipients': msg['To'],
             'recipients_secondary': msg['Cc'],
             'recipients_hidden': bcc,
+            'resource': str(record),
+            'notification': trigger.notification_email.id,
             'trigger': trigger.id,
             }
 
@@ -329,6 +272,24 @@ class Email(ModelSQL, ModelView):
         if logs:
             Log.create(logs)
 
+    @classmethod
+    def validate(cls, notifications):
+        super().validate(notifications)
+        for notification in notifications:
+            notification.check_subject()
+
+    def check_subject(self):
+        if not self.subject:
+            return
+        try:
+            TextTemplate(self.subject)
+        except Exception as exception:
+            raise TemplateError(
+                gettext('notification_email.'
+                    'msg_notification_invalid_subject',
+                    notification=self.rec_name,
+                    exception=exception)) from exception
+
 
 class EmailAttachment(ModelSQL):
     "Email Notification Attachment"
@@ -376,15 +337,17 @@ class EmailAttachment(ModelSQL):
         return msg
 
 
-class EmailLog(ModelSQL, ModelView):
-    "Notitification Email Log"
+class EmailLog(ResourceAccessMixin, ModelSQL, ModelView):
+    "Notification Email Log"
     __name__ = 'notification.email.log'
     date = fields.Function(fields.DateTime('Date'), 'get_date')
     recipients = fields.Char("Recipients")
     recipients_secondary = fields.Char("Secondary Recipients")
     recipients_hidden = fields.Char("Hidden Recipients")
-    trigger = fields.Many2One(
-        'ir.trigger', 'Trigger', required=True, ondelete='CASCADE')
+    notification = fields.Many2One(
+        'notification.email', "Notification",
+        required=True, ondelete='RESTRICT')
+    trigger = fields.Many2One('ir.trigger', "Trigger")
 
     @classmethod
     def __setup__(cls):
@@ -394,6 +357,35 @@ class EmailLog(ModelSQL, ModelView):
             ('id', 'DESC'),
             ]
 
+    @classmethod
+    def __register__(cls, module_name):
+        pool = Pool()
+        Model = pool.get('ir.model')
+        Trigger = pool.get('ir.trigger')
+        model = Model.__table__()
+        trigger = Trigger.__table__()
+        table = cls.__table__()
+        super().__register__(module_name)
+
+        table_h = cls.__table_handler__(module_name)
+        cursor = Transaction().connection.cursor()
+
+        # Migration from 5.6:
+        # fill notification and resource
+        # remove required on trigger
+        notification = trigger.select(
+            trigger.notification_email,
+            where=trigger.id == table.trigger)
+        resource = (trigger
+            .join(model, condition=trigger.model == model.id)
+            .select(
+                Concat(model.model, ',-1'),
+                where=trigger.id == table.trigger))
+        cursor.execute(*table.update(
+                [table.notification, table.resource],
+                [notification, resource]))
+        table_h.not_null_action('trigger', 'remove')
+
     def get_date(self, name):
         return self.create_date.replace(microsecond=0)
 
